-- make car and tank programmable
data.raw.car.car.order = "z["..programmable_identifier.."]"
data.raw.car.tank.order = "z["..programmable_identifier.."]"
data.raw.car.tank.guns = {"tank-cannon"}

-- make all walls more more durable
for _, wall in pairs(data.raw.wall) do
	wall.resistances = wall.resistances or {}
	local resistance_updated = false
	for _2, resistance in ipairs(wall.resistances) do
		if resistance.type == "impact" then
			resistance_updated = true
			resistance.percent = 100 - ((100 - (resistance.percent or 0)) / 2)
			resistance.decrease = (resistance.decrease or 0) + 100
		end
	end
	if not resistance_updated then
		table.insert(wall.resistances, { type = "impact", percent = 50 })
	end
end

function add_resistance(resistances, resistance)
	local resistance_updated = false
	for _2, old_resistance in ipairs(resistances) do
		if old_resistance.type == resistance.type then
			resistance_updated = true
			old_resistance.percent = 100 - ((100 - (old_resistance.percent or 0)) * ((100 - resistance.percent or 0)/100))
			old_resistance.percent = math.max((old_resistance.decrease or 0), (resistance.decrease or 0))
		end
	end
	if not resistance_updated then
		table.insert(resistances, resistance)
	end
end

function make_composite_unit_from_vehicle(vehicle)
	-- do any required vehicle modifications
	if (not vehicle.weight) or vehicle.weight == 0 or string_to_number(vehicle.consumption) < 10 then return end
	
	-- collision_box should be square otherwise it will get stuck during rotation
	local extent_min = math.min(-vehicle.collision_box[1][1], -vehicle.collision_box[1][2], vehicle.collision_box[2][1], vehicle.collision_box[2][2])
	vehicle.collision_box = {{-extent_min,-extent_min},{extent_min,extent_min}}
	vehicle.order = "z["..programmable_identifier.."]" -- make sure it's picked up in control.lua
	vehicle.flags = vehicle.flags or {}
	table.insert(vehicle.flags, "player-creation")
	table.insert(vehicle.flags, "placeable-player")
	vehicle.subgroup = "cars"
	
	-- solid used for auto-vehicle mode base
	local solid = table.deepcopy(vehicle)
	solid.name = solid.name .. composite_suffix.."solid"
	solid.turret_animation = nil
	solid.order = "z-z"
	table.insert(solid.flags, "player-creation")
	solid.localised_name = { "entity-name.".. vehicle.name}
	solid.resistances = solid.resistances or {}
	--add_resistance(solid.resistances, {type = "impact", decrease = 200 }) -- immune to small bumps
	-- immune to bumps breaks the nudge anti-stuck function
	add_resistance(solid.resistances, {type = "impact", percent = 80 }) -- take less damage
	data:extend({solid})
	
	-- ghost used for unit-mode base
	local ghost = table.deepcopy(vehicle)
	ghost.name = ghost.name .. composite_suffix.."ghost"
	ghost.collision_box = {{-0.0, -0.0}, {0.0, 0.0}}
	ghost.collision_mask = { "ghost-layer"}
	ghost.turret_animation = nil
	ghost.order = "z-z"
	table.insert(ghost.flags, "player-creation")
	ghost.localised_name = { "entity-name.".. vehicle.name}
	data:extend({ghost})

	-- turret - a dummy vehicle used as a rotatble graphic for the turret
	if vehicle.turret_animation and vehicle.turret_animation.layers and #vehicle.turret_animation.layers > 0 then 
		local turret_animation = table.deepcopy(vehicle.turret_animation)
		-- offset by inverse of turret y offset so it can be rendered on top
		for _,layer in pairs(turret_animation.layers) do 
			local layer_shift = layer.shift and layer.shift or {0, 0}
			layer_shift[2] = layer_shift[2] - turret_y_offset
			layer.shift = layer_shift
		end
		data:extend({{
			type = "car",
			name = vehicle.name .. composite_suffix.."turret" ,
			icon = vehicle.icon,
			flags = {"pushable", "placeable-neutral", "player-creation", "placeable-off-grid"},
			selectable_in_game = false,
			order="z-z",
			max_health = 1000,
			healing_per_tick = 1000,
			energy_per_hit_point = 1,
			selection_box = {{-0.0, -0.0}, {0.0, 0.0}},
			collision_box = {{-0.0, -0.0}, {0.0, 0.0}},
			collision_mask = { "ghost-layer"},
			terrain_friction_modifier = 1,
			friction = 1,
			animation = turret_animation,
			rotation_speed = 0.01,
			weight = 1000,
			inventory_size = 0,
			consumption = "1W",
			braking_power = "1W",
			burner = {
				effectivity = 1,
				fuel_inventory_size = 1
			},
			effectivity = 1,
			localised_name = { "entity-name.".. vehicle.name}
		}})
	else
		vehicle.turret_animation = nil
	end
	
	-- damage_target - an invisible unit to force enemies to attack this unit
	data:extend({{
		type = "unit",
		name = vehicle.name..composite_suffix.."damage_target",
		icon = vehicle.icon,
		flags = {"player-creation", "placeable-player", "placeable-off-grid"},
		selectable_in_game = false,
		order="z-z",
		max_health = damage_target_max_health,
		healing_per_tick = 0,
		alert_when_damaged = false,
		selection_box = {{0, 0}, {0, 0}},
		collision_box = {{-extent_min,-extent_min},{extent_min,extent_min}},
		--collision_mask = { "item-layer", "object-layer", "player-layer", "water-tile"},
		collision_mask = { "doodad-layer"},
		vision_distance = 1,
		movement_speed = 0,
		pollution_to_join_attack = 0.0,
		distraction_cooldown = 0,
		distance_per_frame =  0,
		resistances = vehicle.resistances,
		run_animation = {
			filename = "__aai-programmable-vehicles__/graphics/blank.png",
			width = 1,
			height = 1,
			frame_count = 1,
			direction_count = 1,
		},
		attack_parameters = {
			ammo_category = "melee",
			ammo_type = {
			  action = {
				action_delivery = {
				  target_effects = {
					damage = {
					  amount = 5, -- the damage needs to be there to simulate drive-crushing damage otherwise unit gets stuck
					  type = "impact"
					},
					type = "damage"
				  },
				  type = "instant"
				},
				type = "direct"
			  },
			  category = "melee",
			  target_type = "entity"
			},
			animation = {
                filename = "__aai-programmable-vehicles__/graphics/blank.png", 
                width = 1, 
                height = 1 ,
                frame_count = 1,
                direction_count = 1,
            },
			cooldown = 30,
			range = 0.5,
			type = "projectile"
		},
		localised_name = { "entity-name.".. vehicle.name}
	}})
	
	local show_navigator = true
	-- navigator - an invisible unit to direct the vehicle while in unit (move_to) mode
	data:extend({{
		type = "unit",
		name = vehicle.name..composite_suffix.."navigator",
		icon = vehicle.icon,
		flags = {"placeable-neutral", "placeable-off-grid"},
		selectable_in_game = false,
		order="z-z",
		max_health = 1000000,
		healing_per_tick = 1000000,
		alert_when_damaged = false,
		selection_box = {{-1.0, -1.0}, {1.0, 1.0}},
		collision_box = {{-extent_min,-extent_min},{extent_min,extent_min}},
		--collision_mask = { "item-layer", "object-layer", "player-layer", "water-tile"},
		--collision_mask = { "item-layer", "object-layer", "water-tile"},
		vision_distance = 1,
		movement_speed = 1 + math.sqrt(string_to_number(vehicle.consumption) * vehicle.effectivity / vehicle.weight) / 50, -- should be faster than vehicle speed. The vehicle holds it back
		pollution_to_join_attack = 0.0,
		distraction_cooldown = 0,
		distance_per_frame =  0.05,
		run_animation = (show_navigator and {
			filename = "__aai-programmable-vehicles__/graphics/entity/debug-navigator.png",
			width = 128,
			height = 128,
			frame_count = 1,
			direction_count = 1,
		} or {
			filename = "__aai-programmable-vehicles__/graphics/blank.png",
			width = 1,
			height = 1,
			frame_count = 1,
			direction_count = 1,
		}),
		attack_parameters = {
			ammo_category = "melee",
			ammo_type = {
			  action = {
				action_delivery = {
				  target_effects = {
					damage = {
					  amount = 5, -- the damage needs to be there to simulate drive-crushing damage otherwise unit gets stuck
					  type = "impact"
					},
					type = "damage"
				  },
				  type = "instant"
				},
				type = "direct"
			  },
			  category = "melee",
			  target_type = "entity"
			},
			animation = {
                filename = "__aai-programmable-vehicles__/graphics/blank.png", 
                width = 1, 
                height = 1 ,
                frame_count = 1,
                direction_count = 1,
            },
			cooldown = 30,
			range = 0.5,
			type = "projectile"
		},
		localised_name = { "entity-name.".. vehicle.name}
	}})
	
	-- buffer - an invisible object used to find open spaces and avoid getting stuck on things
	local buffer = extent_min * 1.5
	data:extend({{
		type = "decorative",
		name = vehicle.name..composite_suffix.."buffer",
		icon = vehicle.icon,
		flags = {"placeable-neutral", "placeable-off-grid"},
		subgroup = "grass",
		order = "z-z",
		collision_box = {{-buffer, -buffer}, {buffer, buffer}},
		collision_mask = vehicle.collision_mask or { "item-layer", "player-layer", "water-tile"},
		selection_box = {{-0.0, -0.0}, {0.0, 0.0}},
		selectable_in_game = false,
		render_layer = "resource",
		pictures = {{ 
            filename = "__aai-programmable-vehicles__/graphics/blank.png", 
            width = 1, 
            height = 1 
        }},
		localised_name = { "entity-name.".. vehicle.name}
	}})
	
	local lamp = {
		type = "fire",
		name = vehicle.name..composite_suffix.."lamp",
		spawn_entity = "fire-flame-on-tree",
		damage_per_tick = { amount = 0, type="poison"},
		emissions_per_tick = 0.005,
		flags = { "placeable-off-grid", "not-on-map" },
		flame_alpha = 0.35,
		flame_alpha_deviation = 0.05,
		light = { intensity = 0.1, size = 40 },
        start_scale = 0.2,
		initial_lifetime = 480,
        duration = 480,
        lifetime_increase_by = 20,
        lifetime_increase_cooldown = 10,
        increase_duration_by = 20,
        increase_duration_cooldown = 10,
		delay_between_initial_flames = 60,
		fade_in_duration = 0,
		fade_out_duration = 0,
        fade_away_duration = 0,
		maximum_spread_count = 1,
		smoke_fade_in_duration = 0,
		smoke_fade_out_duration = 0,
		spread_delay = 0,
		spread_delay_deviation = 0,
        spread_duration = 0,
		tree_dying_factor = 0,
		pictures = {
			{
				filename = "__aai-programmable-vehicles__/graphics/blank.png",
				width = 1,
				height = 1,
				line_length = 1,
				frame_count = 1,
				axially_symmetrical = false,
				blend_mode = "additive",
				flags = {
					"compressed"
				},
				direction_count = 1,
			},
		},
		small_tree_fire_pictures = {
			{
				filename = "__aai-programmable-vehicles__/graphics/blank.png",
				width = 1,
				height = 1,
				line_length = 1,
				frame_count = 1,
				axially_symmetrical = false,
				blend_mode = "additive",
				flags = {
					"compressed"
				},
				direction_count = 1,
			},
		},
		smoke_source_pictures = {
			{
				filename = "__aai-programmable-vehicles__/graphics/blank.png",
				width = 1,
				height = 1,
				line_length = 1,
				frame_count = 1,
				axially_symmetrical = false,
				blend_mode = "additive",
				flags = {
					"compressed"
				},
				direction_count = 1,
			},
		}
    }
	
	local lamp_sound = table.deepcopy(lamp)
	lamp_sound.name = vehicle.name..composite_suffix.."lamp_sound"
	lamp_sound.light = { intensity = 0.2, size = 40 }
	--[[working_sound =
    {
      sound =
      {
        filename = "__base__/sound/car-engine.ogg",
        volume = 0.6
      },
      activate_sound =
      {
        filename = "__base__/sound/car-engine-start.ogg",
        volume = 0.6
      },
      deactivate_sound =
      {
        filename = "__base__/sound/car-engine-stop.ogg",
        volume = 0.6
      },
      match_speed_to_activity = true,
    },]]--
	-- TODO: add activate_sound and deactivate_sound as explosions
	if vehicle.working_sound and (vehicle.working_sound.filename or vehicle.working_sound.sound) then
		lamp_sound.working_sound = {sound = table.deepcopy(vehicle.working_sound.sound or vehicle.working_sound)}
		-- increase volume to compensate for distance
		lamp_sound.working_sound.sound.volume = (lamp_sound.working_sound.volume or 1) * 1.1 
		lamp_sound.working_sound.max_sounds_per_type = 3
	end
	
	data:extend({lamp, lamp_sound})
	
	
	
	
	-- virtual signal: allows unit to be specified without confusion with inventory items (unit items)
	local virtual_signal = {
			type = "virtual-signal",
			name = vehicle.name..composite_suffix.."signal",
			--icon = vehicle.icon,
			icons = {
				{
				  icon = "__aai-programmable-vehicles__/graphics/icons/ids/id_background.png"
				},
				{
				  icon = vehicle.icon
				},
				{
				  icon = "__aai-programmable-vehicles__/graphics/icons/ids/id_overlay.png"
				},
			},
			subgroup = "vehicle-ids",
			order = "a[vehicle]",
			localised_name = { "vehicle-signal", {"entity-name.".. vehicle.name}}
	}
	data:extend({ virtual_signal })
	
end

-- programmable-warfare vehcile/units
local vehicles_to_process = {}
for _,vehicle in pairs(data.raw.car) do -- beware loop
	--if vehicle.order and string.find(vehicle.order, programmable_identifier, 1, true) then 
	if not string.find(vehicle.name, composite_suffix, 1, true) then 
		table.insert(vehicles_to_process, vehicle)
	end
end

for _,vehicle in pairs(vehicles_to_process) do
	make_composite_unit_from_vehicle(vehicle)
end